import pandas as pd
import numpy as np
import datetime
import traceback
from ...api.raw import *
from ....util.calculator_item import *
from ...wrapper.mysql import ViewDatabaseConnector
from ...view.view_models import StockIpoCollection, ConvBondIpoCollection

class StockIpoCollectionProcessor(object):

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def calc(self):
        col_dic = {
            'IPOPRICE':'发行价',
            'IPOISSUEDATE':'首发发行日期',
            'IPOSHARESVOL':'首发数量',
            'IPOPE':'发行市盈率',
            'NAME':'股票简称',
            'SUCRATIOONL':'中签率',
            'PURCHACODEONL':'网上申购代码带市场',
            'CEILINGONL':'申购上限',
            'EMIND2016':'东财行业',
            'TRADEMARKET':'上市板块',
            'IPOPURCHDATEONL':'申购日期',
            'IPOANNCDATE':'发行公告日',
            'WSZQJGGGDATE':'中签公布',
            'LISTDATE':'上市日期',
            'CODES':'股票代码',
        }  
        try:
            with RawDatabaseConnector().managed_session() as db_session:
                query = db_session.query(EmStockIpoInfo)
                df = pd.read_sql(query.statement, query.session.bind)
            stock_list = df.CODES.tolist()
            df = df.rename(columns=col_dic) 
            stock_close = RawDataApi().get_em_stock_price(stock_list=stock_list)
            stock_close = stock_close.pivot_table(index='datetime',columns='stock_id',values='close')
            res = []
            for r in df.itertuples():
                stock_code = r.股票代码
                dates = stock_close.index.values
                if stock_code not in stock_close.columns:
                    continue
                values = stock_close[stock_code].values
                ipo_price = r.发行价
                list_date = r.上市日期
                ipo_date = r.首发发行日期
                trade_market = r.上市板块
                stock_name = r.股票简称
                res_status = CalculatorBase.stock_ipo_continue_up_limit(dates=dates,
                                    values=values,
                                    ipo_price=ipo_price,
                                    list_date=list_date,
                                    ipo_date=ipo_date,
                                    trade_market=trade_market,
                                    stock_name=stock_name,
                                    stock_code=stock_code)
                res.append(res_status)
            self.result = pd.DataFrame(res)
            self.result.replace([np.Inf,-np.Inf],np.nan,regex=True, inplace = True)
            self.append_data(StockIpoCollection.__tablename__, self.result)
            return True

        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def process(self):
        failed_tasks = []
        if not self.calc():
            failed_tasks.append('collection_stock_ipo')
        return failed_tasks


class ConcBondIpoCollectionProcessor(object):

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def calc(self):
        col_dic = {
            'CODES':'转债代码',
            'CBCODE' : '转债代码无市场',
            'CBNAME' : '转债简称',
            'CBSTOCKCODE':'正股代码',
            'CBSTOCKNAME':'正股简称',
            'CBISSUEAMT':'发行总额(亿)',
            'IPOANNCDATE':'发行公告日',
            'CBIPOTYPE':'发行方式',
            'CBPURCODEONL':'网上发行申购代码',
            'ISSRATE':'发行时债项评级',
            'CBTERM':'发行期限',
            'CBISSUEPRICE':'发行价格',
            'ISSUEONLEDATE':'上网发行截止日期',
            'CBCONVPRICE':'转股价',
            'CBDATEONL':'上网发行日期',
            'CBLISTDATE':'上市日期',
            'CBRESULTDATE':'中签公布',
            'CBSUCRATEONL':'中签率',
            'QUANTLIMITINTRO':'网上发行认购数量限制说明',   
        }
        try:
            with RawDatabaseConnector().managed_session() as db_session:
                query = db_session.query(EmConvBondIpoInfo)
                df = pd.read_sql(query.statement, query.session.bind)
            df = df.rename(columns=col_dic)
            conv_bond_list = df.转债代码.tolist()
            data = RawDataApi().get_em_conv_bond_dirty_price(codes=conv_bond_list)
            conv_bond_close = data.pivot_table(index='datetime',columns='bond_id',values='close')
            conv_bond_open = data.pivot_table(index='datetime',columns='bond_id',values='open')
            res = []
            for r in df.itertuples():
                bond_id = r.转债代码
                bond_name = r.转债简称
                list_date = r.上市日期
                ipo_price = r.发行价格
                dates = conv_bond_close.index.values
                if bond_id not in conv_bond_close.columns.tolist():
                    continue
                if r.发行方式 == '私募':
                    continue
                close_values = conv_bond_close[bond_id].values
                open_values = conv_bond_open[bond_id].values
                rate = r.中签率
                limit_num = r.网上发行认购数量限制说明
                res_i = CalculatorBase.conv_bond_ipo_result(bond_id=bond_id,
                                                    bond_name=bond_name,
                                                    list_date=list_date,
                                                    ipo_price=ipo_price,
                                                    dates=dates,
                                                    close_values=close_values,
                                                    open_values=open_values,
                                                    rate=rate,
                                                    limit_num=limit_num,
                                                )
                res.append(res_i)
            self.result = pd.DataFrame(res)
            self.result.replace([np.Inf,-np.Inf],np.nan,regex=True,inplace=True)
            self.append_data(ConvBondIpoCollection.__tablename__, self.result)
            return True

        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def process(self):
        failed_tasks = []
        if not self.calc():
            failed_tasks.append('collection_conv_bond_ipo')
        return failed_tasks


if __name__ == '__main__':
    ConcBondIpoCollectionProcessor().process()



from .fund_derived_factors import *
from ...data.fund.derived.derived_data_helper import normalize, score_rescale

class ScoreCalculator:

    @staticmethod
    def fund_ability_factor_prepare(factor_list, wind_type_dict):
        dt = datetime.date(2012,1,1)
        result = []
        for fac_name in factor_list:
            obj = eval(fac_name)()
            fac = obj.get()
            fac = fac[fac.index > dt]
            fac.columns = pd.MultiIndex.from_product([[fac_name], fac.columns])
            result.append(fac)
            obj.clear(recursive=True)
        _factor = pd.concat(result, axis=1).stack()
        _factor = _factor.reset_index().dropna(subset=factor_list, how='all')
        fund_info = FundInfo().get()
        fund_type = ['stock','bond','index','QDII','mmf']
        wind_class_type = fund_info.set_index('fund_id')['wind_class_2']
        _factor = _factor.rename(columns={'level_1':'fund_id'})
        _factor.loc[:,'fund_type'] = _factor.fund_id.map(lambda x : wind_type_dict.get(wind_class_type.get(x),'stock'))
        _factor = _factor.set_index('datetime')
        return _factor

    @staticmethod
    def score_rescale(df):
        return df.rank(pct=True) * 100
    
    @staticmethod
    def process_score(df):
        df = df.reset_index().set_index('fund_id').drop(columns=['fund_type'])
        return df.rank(pct=True) * 100
    
    @staticmethod
    def calc_fund(df, fund_type, dt, calcu_score_fund):
        df = ScoreCalculator.process_score(df)
        if df is None:
            return None
        df = calcu_score_fund(df, fund_type)
        if df is None:
            return None
        df.loc[:,'fund_type'] = fund_type
        df.loc[:,'datetime'] = dt
        return df.dropna(subset=['fund_score'])

    @staticmethod
    def ability_calc(_factor, calcu_score_fund):
        date_list = sorted(_factor.index.unique().tolist())
        res = []
        lens_dt = len(date_list)
        fund_types = ['stock','bond','index','QDII','mmf']
        _t0 = time.time()
        for dt in date_list:
            df_dt = _factor.loc[dt].set_index('fund_type').copy()
            for fund_type in fund_types:
                try:
                    df_dt_fund_type = df_dt.loc[fund_type]
                except:
                    continue
                res.append(ScoreCalculator.calc_fund(df_dt_fund_type, fund_type, dt, calcu_score_fund))
            dt_idx = date_list.index(dt)
            if dt_idx % 100 == 0:
                _t1 = time.time()
                print(f'\t\tdt {dt} idx {dt_idx} total {lens_dt} cost {round(_t1 - _t0)}')
                _t0 = time.time()
        res = [i for i in res if i is not None]
        _factor = pd.concat(res, axis=0)
        _factor = _factor.reset_index().pivot_table(index='datetime',values='fund_score',columns='fund_id')
        return _factor

class MngScoreV1(Factor):
    # 雷达图基金评价 历史业绩为重

    def __init__(self):
        super().__init__(f_name='MngScoreV1', f_type=FundFactorType.DERIVED, f_level='score')
        self._deps.add(MngAnnualRetDailyHistory())
        self._deps.add(MngTotalRetDailyHistory())
        self._deps.add(MngMddDailyHistory())
        self._deps.add(MngAnnualVolDailyHistory())
        self._deps.add(MngDownsideStdDailyHistory())
        self._deps.add(MngFundTypeTradingDays())
        self._deps.add(MngFundSize())
        self._deps.add(MngClAlphaWeeklyHistory())
        self._deps.add(MngClBetaWeeklyHistory())

    def calc_mng(self, df, fund_type):
        if fund_type == 'stock':
            return self.score_mng_stock(df)
        if fund_type == 'bond':
            return self.score_mng_bond(df)
        if fund_type == 'index':
            return self.score_mng_index(df)
        if fund_type == 'QDII':
            return self.score_mng_qdii(df)
        if fund_type == 'mmf':
            return self.score_mng_mmf(df)
        
    def score_mng_stock(self, df):
        df.loc[:,'mng_ret_ability'] = score_rescale(0.4 * score_rescale(df['MngAnnualRetDailyHistory']) +0.6 * score_rescale(df['MngTotalRetDailyHistory']))
        df.loc[:,'mng_risk_ability'] = score_rescale(0.3 * score_rescale(-df['MngMddDailyHistory']) + 0.4 * score_rescale(-df['MngAnnualVolDailyHistory']) + 0.3 * score_rescale(-df['MngDownsideStdDailyHistory']))
        df.loc[:,'mng_select_stock'] =  score_rescale(normalize(df['MngClAlphaWeeklyHistory']))
        df.loc[:,'mng_select_time'] = score_rescale(normalize(df['MngClBetaWeeklyHistory']) )
        df.loc[:,'mng_experience'] = score_rescale( 0.8 *  score_rescale(df['MngFundTypeTradingDays']) + 0.2 * score_rescale(df['MngFundSize']))
        df.loc[:,'mng_score'] = score_rescale(df.loc[:,'mng_ret_ability'] + df.loc[:,'mng_risk_ability'] + df.loc[:,'mng_select_time'] + df.loc[:,'mng_select_stock'] + 1.5 * df.loc[:,'mng_experience'])
        return df[['mng_id','mng_ret_ability','mng_risk_ability','mng_select_time','mng_select_stock','mng_experience','mng_score']]

    def score_mng_bond(self, df):
        df.loc[:,'mng_ret_ability'] = score_rescale( 0.8 * score_rescale(df['MngAnnualRetDailyHistory']) + 0.2 * score_rescale(df['MngTotalRetDailyHistory']))
        df.loc[:,'mng_risk_ability'] = score_rescale(0.1 * score_rescale(-df['MngMddDailyHistory']) + 0.4 * score_rescale(-df['MngAnnualVolDailyHistory']) + 0.5 * score_rescale(-df['MngDownsideStdDailyHistory']))
        df.loc[:,'mng_select_stock'] =  score_rescale(normalize(df['MngClAlphaWeeklyHistory']))
        df.loc[:,'mng_select_time'] = score_rescale(normalize(df['MngClBetaWeeklyHistory']))
        df.loc[:,'mng_experience'] = score_rescale( 0.8 *  score_rescale(df['MngFundTypeTradingDays']) + 0.2 * score_rescale(df['MngFundSize']))
        df.loc[:,'mng_score'] = score_rescale(1.2 * df.loc[:,'mng_ret_ability'] +  df.loc[:,'mng_risk_ability'] + df.loc[:,'mng_select_time'] + df.loc[:,'mng_select_stock'] + 1.5 * df.loc[:,'mng_experience'])
        return df[['mng_id','mng_ret_ability','mng_risk_ability','mng_select_time','mng_select_stock','mng_experience','mng_score']]

    def score_mng_index(self, df):
        df.loc[:,'mng_ret_ability'] = score_rescale( 0.4 * score_rescale(df['MngAnnualRetDailyHistory']) + 0.6 * score_rescale(df['MngTotalRetDailyHistory']))
        df.loc[:,'mng_risk_ability'] = score_rescale(0.3 * score_rescale(-df['MngMddDailyHistory']) + 0.4 * score_rescale(-df['MngAnnualVolDailyHistory']) + 0.3 * score_rescale(-df['MngDownsideStdDailyHistory']))
        df.loc[:,'mng_experience'] = score_rescale( 0.4 *  score_rescale(df['MngFundTypeTradingDays']) + 0.6 * score_rescale(df['MngFundSize']))
        df.loc[:,'mng_score'] = score_rescale(df.loc[:,'mng_ret_ability'] + df.loc[:,'mng_risk_ability'] + 2 * df.loc[:,'mng_experience'] )
        return df[['mng_id','mng_ret_ability','mng_risk_ability','mng_experience','mng_score']]

    def score_mng_qdii(self, df):
        df.loc[:,'mng_ret_ability'] = score_rescale( 0.4 * score_rescale(df['MngAnnualRetDailyHistory']) + 0.6 * score_rescale(df['MngTotalRetDailyHistory']))
        df.loc[:,'mng_risk_ability'] = score_rescale(0.3 * score_rescale(-df['MngMddDailyHistory']) + 0.4 * score_rescale(-df['MngAnnualVolDailyHistory']) + 0.3 * score_rescale(-df['MngDownsideStdDailyHistory']))
        df.loc[:,'mng_experience'] = score_rescale( 0.8 *  score_rescale(df['MngFundTypeTradingDays']) + 0.2 * score_rescale(df['MngFundSize']))
        df.loc[:,'mng_score'] = score_rescale(df.loc[:,'mng_ret_ability'] + df.loc[:,'mng_risk_ability'] + 1.5 * df.loc[:,'mng_experience'])
        return df[['mng_id','mng_ret_ability','mng_risk_ability','mng_experience','mng_score']]

    def score_mng_mmf(self, df):
        df.loc[:,'mng_ret_ability'] = score_rescale(normalize(df['MngAnnualRetDailyHistory']))
        df.loc[:,'mng_experience'] = score_rescale( 0.8 *  score_rescale(df['MngFundTypeTradingDays']) + 0.2 * score_rescale(df['MngFundSize']))
        df.loc[:,'mng_score'] = score_rescale(df.loc[:,'mng_ret_ability'] + df.loc[:,'mng_experience'])
        return df[['mng_id','mng_ret_ability','mng_experience','mng_score']]

    def prepare(self):
        dt = datetime.date(2012,1,1)
        factor_list = [i.name for i in self._deps]
        result = []
        for fac_name in factor_list:
            obj = eval(fac_name)()
            fac = obj.get()
            fac = fac[fac.index > dt]
            fac.columns = pd.MultiIndex.from_product([[fac_name], fac.columns])
            result.append(fac)
            #obj.clear(recursive=True)
        _factor = pd.concat(result, axis=1).stack()
        _factor = _factor.reset_index().dropna(subset=factor_list, how='all').rename(columns={'level_1':'mng_id'})
        _factor.loc[:,'fund_type'] = _factor.mng_id.map(lambda x: x.split('_')[0])
        _factor = _factor.set_index(['datetime','fund_type'])
        return _factor

    def calc(self):
        _factor = self.prepare()
        fund_types = ['stock','bond','index','QDII','mmf']
        date_list = sorted(_factor.index.get_level_values(0).unique().tolist())
        result = []
        _t0 = time.time()
        for dt in date_list:
            #df_dt = self._factor.loc[dt]
            for fund_type in fund_types:
                df_i = _factor.loc[dt,fund_type].copy()
                df_i = self.calc_mng(df_i, fund_type)
                result.append(df_i)
            dt_idx = date_list.index(dt)
            if dt_idx % 100 == 0:
                print(f'dt {dt} idx {dt_idx}')
        _t1 = time.time()
        print(_t1 - _t0)
        self._factor = pd.concat(result).dropna(subset=['mng_score'])
        self._factor = self._factor.reset_index()
        self._factor = self._factor.set_index('datetime')
        td = self._factor.index.tolist()
        td = pd.to_datetime(td)
        td = [i.date() for i in td] 
        self._factor.index = td

    def append_update(self):
        _exsited_factor = self.get()
        fund_types = ['stock','bond','index','QDII','mmf']
        last_date = pd.to_datetime(_exsited_factor.index[-1]).date()
        _factor = self.prepare()
        date_list = sorted(_factor.index.get_level_values(0).unique().tolist())
        date_list = [pd.to_datetime(i).date() for i in date_list]
        date_list = [i for i in date_list if i > last_date]
        result = []
        if len(date_list)  == 0:
            return True
        for dt in date_list:
            for fund_type in fund_types:
                df_i = _factor.loc[dt,fund_type].copy()
                df_i = self.calc_mng(df_i, fund_type)
                result.append(df_i)
            dt_idx = date_list.index(dt)
        if len(result) == 0:
            return True
        self._factor = pd.concat(result).dropna(subset=['mng_score'])
        self._factor = self._factor.reset_index().set_index('datetime')
        self._factor = _exsited_factor.append(self._factor)
        td = self._factor.index.tolist()
        td = pd.to_datetime(td)
        td = [i.date() for i in td] 
        self._factor.index = td
        return self.save()

class FundMngScoreV1(Factor):

    def __init__(self):
        super().__init__(f_name='FundMngScoreV1', f_type=FundFactorType.DERIVED, f_level='score')
        self._deps.add(MngScoreV1())
        self._deps.add(FundManagerInfo())
        self._deps.add(FundNavDaily())
        self._deps.add(FundInfo())
   
    def prepare(self):
        total_score = MngScoreV1().get().reset_index()
        fund_manager_info = FundManagerInfo().get().set_index('fund_id')[['mng_id','start_date','end_date']]
        fund_nav = FundNavDaily().get()
        fund_info = FundInfo().get()
        type_list = ['stock','bond','index','QDII','mmf']
        manager_score = {}
        for type_i in type_list:
            _df = total_score[total_score['fund_type'] == type_i].reset_index().rename(columns={'index':'datetime'})
            _df = _df.pivot_table(index='datetime',columns='mng_id',values='mng_score')
            manager_score[type_i] = _df
        date_index = manager_score['stock'].index
        #date_index = [_.date() for _ in date_index]
        wind_class_dict = fund_info.set_index('fund_id').to_dict()['wind_class_2']
        return fund_manager_info, fund_nav, wind_class_dict, manager_score, date_index

    def loop_item(self, fund_id, wind_class_dict, fund_manager_info, manager_score, date_index):
        wind_type = wind_class_dict.get(fund_id)
        fund_type = Calculator.WIND_TYPE_DICT.get(wind_type, 'stock')
        if fund_id not in fund_manager_info.index:
            return None
        single_fund_info = fund_manager_info.loc[[fund_id]]
        result = []
        for row in single_fund_info.itertuples(index=False):
            mng_id = fund_type + '_' + row.mng_id
            if mng_id not in manager_score[fund_type]:
                continue
            data = (date_index >= row.start_date) & (date_index <= row.end_date)
            _manager_score = pd.Series(1, index=date_index[data], name=mng_id)
            _manager_score = _manager_score * manager_score[fund_type][mng_id]
            result.append(_manager_score)
        if result  == []:
            return None
        fund_manager_score_i = pd.concat(result,axis=1).max(axis=1)
        fund_manager_score_i.name = fund_id
        return fund_manager_score_i

    def calc(self):
        fund_manager_info, fund_nav, wind_class_dict, manager_score, date_index = self.prepare()
        score_result_total = []
        fund_list = fund_nav.columns.tolist()
        _t0 = time.time()
        for fund_id in fund_list:
            res_i = self.loop_item(fund_id, wind_class_dict, fund_manager_info, manager_score, date_index)
            if res_i is not None:
                score_result_total.append(res_i)
            idx = fund_list.index(fund_id)
            if idx % 1000 == 0:
                _t1 = time.time()
                #print(f'fund_id {fund_id} {idx} {len(fund_list)} cost { round(_t1 - _t0,2)}')
                _t0 = time.time()
        self._factor = pd.concat(score_result_total, axis=1)

class RetAbilityFundScore(Factor):

    def __init__(self):
        super().__init__(f_name='RetAbilityFundScore', f_type=FundFactorType.DERIVED, f_level='score')
        self._deps.add(AnnualRetDailyHistory())
        self._deps.add(TotalRetDailyHistory())
        self._deps.add(RecentMonthRet())

    @staticmethod
    def calcu_score_fund(df, fund_type):
        if fund_type in ['stock','index','QDII']:
            df.loc[:,'fund_score'] = ScoreCalculator.score_rescale(0.4 * df.AnnualRetDailyHistory + 0.6 * df.TotalRetDailyHistory)
        elif fund_type in ['bond']:
            df.loc[:,'fund_score'] = ScoreCalculator.score_rescale(0.8 * df.AnnualRetDailyHistory + 0.2 * df.TotalRetDailyHistory)
        elif fund_type in ['mmf']:
            df.loc[:,'fund_score'] = ScoreCalculator.score_rescale(df.RecentMonthRet)
        return df[['fund_score']]

    def calc(self):
        fund_list = [i.name for i in self._deps]
        factor = ScoreCalculator.fund_ability_factor_prepare(fund_list, Calculator.WIND_TYPE_DICT)
        self._factor = ScoreCalculator.ability_calc(factor, RetAbilityFundScore.calcu_score_fund)

class RiskAbilityFundScore(Factor):

    def __init__(self):
        super().__init__(f_name='RiskAbilityFundScore', f_type=FundFactorType.DERIVED, f_level='score')
        self._deps.add(MddDailyHistory())
        self._deps.add(AnnualVolDailyHistory())
        self._deps.add(DownsideStdDailyHistory())
        self._deps.add(FundPersonalHold())
        self._deps.add(FundSizeCombine())
        
    @staticmethod
    def calcu_score_fund(df, fund_type):
        if fund_type in ['stock','index','QDII','bond']:
            df.loc[:,'fund_score'] = ScoreCalculator.score_rescale(0.3 * df.MddDailyHistory + 0.4 * df.AnnualVolDailyHistory + 0.3 * df.DownsideStdDailyHistory)
        elif fund_type in ['mmf']:
            df.loc[:,'fund_score'] = ScoreCalculator.score_rescale(0.5 * df.FundPersonalHold + 0.5 * df.FundSizeCombine)
        return df[['fund_score']]

    def calc(self):
        fund_list = [i.name for i in self._deps]
        factor = ScoreCalculator.fund_ability_factor_prepare(fund_list, Calculator.WIND_TYPE_DICT)
        self._factor = ScoreCalculator.ability_calc(factor, RiskAbilityFundScore.calcu_score_fund)

class StableAbility(Factor):

    def __init__(self):
        super().__init__(f_name='StableAbility', f_type=FundFactorType.DERIVED, f_level='score')
        self._deps.add(TradeYear())
        self._deps.add(ContinueRegValue())
        self._deps.add(FundSizeCombine())

    @staticmethod
    def calcu_score_fund(df, fund_type):
        if fund_type in ['stock','index','QDII','bond']:
            df.loc[:,'fund_score'] = ScoreCalculator.score_rescale(0.4 * df.TradeYear + 0.4 * df.ContinueRegValue + 0.2 * df.FundSizeCombine)
        else:
            return None
        return df[['fund_score']]

    def calc(self):
        fund_list = [i.name for i in self._deps]
        factor = ScoreCalculator.fund_ability_factor_prepare(fund_list, Calculator.WIND_TYPE_DICT)
        self._factor = ScoreCalculator.ability_calc(factor, StableAbility.calcu_score_fund)

class SelectTimeAbility(Factor):

    def __init__(self):
        super().__init__(f_name='SelectTimeAbility', f_type=FundFactorType.DERIVED, f_level='score')
        self._deps.add(FundClBetaHistoryWeekly())
    
    @staticmethod
    def calcu_score_fund(df, fund_type):
        if fund_type in ['stock','QDII','index']:
            df.loc[:,'fund_score'] = ScoreCalculator.score_rescale(df.FundClBetaHistoryWeekly)
        else:
            return None
        return df[['fund_score']]

    def calc(self):
        fund_list = [i.name for i in self._deps]
        factor = ScoreCalculator.fund_ability_factor_prepare(fund_list, Calculator.WIND_TYPE_DICT)
        self._factor = ScoreCalculator.ability_calc(factor, SelectTimeAbility.calcu_score_fund)

class SelectStockAbility(Factor):

    def __init__(self):
        super().__init__(f_name='SelectStockAbility', f_type=FundFactorType.DERIVED, f_level='score')
        self._deps.add(FundClAlphaHistoryWeekly())

    @staticmethod
    def calcu_score_fund(df, fund_type):
        if fund_type in ['stock','QDII','index']:
            df.loc[:,'fund_score'] = ScoreCalculator.score_rescale(df.FundClAlphaHistoryWeekly)
        else:
            return None
        return df[['fund_score']]

    def calc(self):
        fund_list = [i.name for i in self._deps]
        factor = ScoreCalculator.fund_ability_factor_prepare(fund_list, Calculator.WIND_TYPE_DICT)
        self._factor = ScoreCalculator.ability_calc(factor, SelectStockAbility.calcu_score_fund)

class UpdateDerivedScoreStart:
    pass




from typing import Tuple, List, Dict
from functools import partial
import math
import datetime
from dataclasses import dataclass, field
import numpy as np
import pandas as pd

from ...api.raw import RawDataApi
# from ...view.view_models import TianData
# from ...wrapper.mysql import ViewDatabaseConnector


@dataclass
class TianCollectionProcessor():
    # 一些常量
    YI: float = field(init=False, default=math.pow(10, 8))
    FIN_DATE_LIST: Tuple[str] = field(init=False, default=('2017-12-31', '2018-12-31', '2019-12-31'))

    # 内部使用，非参数
    _industry_rank: Dict[str, pd.DataFrame] = field(init=False, default_factory=dict)
    _mb_industry_rank: Dict[str, pd.DataFrame] = field(init=False, default_factory=dict)

    def _scored_by_valuation(self, industry_df: pd.DataFrame, value_df: pd.Series) -> int:
        # 权重
        w = [0.6, 0.3, 0.1]
        weight: float = self._industry_rank[value_df.bl_em_ind_code].loc[value_df.name, ['pe_ttm_deducted', 'pb_lyr_n', 'ps_ttm']].mul(w).sum()

        industry_value = industry_df.loc[value_df.bl_em_ind_code, :]
        gt_pe = value_df.pe_ttm_deducted > industry_value.pe_ttm_deducted
        gt_pb = value_df.pb_lyr_n > industry_value.pb_lyr_n
        gt_ps = value_df.ps_ttm > industry_value.ps_ttm
        value_sum = gt_pe + gt_pb + gt_ps
        if value_sum == 3:
            return round(1 + 2 * weight)
        elif value_sum == 2:
            return round(4 + weight)
        elif value_sum == 1:
            return round(6 + weight)
        else:
            return round(8 + 2 * weight)

    def _scored_by_pledge_ratio(self, value_df: pd.Series) -> int:
        value = value_df.hold_frozen_amt_accum_ratio
        if np.isnan(value):
            return 10
        elif value >= 70:
            return 4
        elif value >= 60:
            return 5
        elif value >= 50:
            return 6
        elif value >= 40:
            return 7
        elif value >= 30:
            return 8
        else:
            return 10

    def _scored_by_price(self, price_df: pd.DataFrame, value_df: pd.DataFrame) -> int:
        score = 5
        if value_df.close >= price_df[price_df.loc[:, 'stock_id'] == value_df.name].loc[:, 'close'].quantile(q=0.8):
            score -= 1
        return score

    def _scored_by_industry(self, value_df: pd.DataFrame) -> float:
        # 权重
        w = [0.4, 0.4, 0.1, 0.05, 0.05]
        try:
            weight: float = self._mb_industry_rank[value_df.mb].loc[value_df.name, ['mv', 'total_asset', 'balance_statement_141_2019-12-31', 'income_statement_9_2019-12-31', 'income_statement_61_2019-12-31']].mul(w).sum()
        except KeyError:
            weight: float = self._industry_rank[value_df.bl_em_ind_code].loc[value_df.name, ['mv', 'total_asset', 'balance_statement_141_2019-12-31', 'income_statement_9_2019-12-31', 'income_statement_61_2019-12-31']].mul(w).sum()
        return round(weight * 10, 2)

    def _get_mb(self, value_df: pd.Series) -> str:
        mb_sales_cons = value_df['mb_sales_cons_2019-12-31'] or value_df['mb_sales_cons_2018-12-31'] or value_df['mb_sales_cons_2017-12-31']
        if pd.isnull(mb_sales_cons):
            return ''
        # 简单取主营收入构成中的第一项的名称
        first_mb = mb_sales_cons.split(',')[0]
        return first_mb.split(':')[0]

    # rzw TODO: 日期参数化
    # rzw TODO: 不要在列后边加日期，使用组合索引(日期也作为索引之一)
    def compute_collection(self, end_date: str):
        raw_data_api = RawDataApi()
        # 取18个月的股价
        price_df = raw_data_api.get_em_stock_price('2018-09-01', end_date)
        # 股票基本信息
        info_df = raw_data_api.get_em_stock_info().set_index('stock_id')
        # 股票每日更新数据，基本只取最新一天的就可以
        daily_df = raw_data_api.get_em_daily_info(start_date=end_date, end_date=end_date).set_index('stock_id')
        # 股票财务数据（季报，年报数据）
        fin_df = raw_data_api.get_em_stock_fin_fac(date_list=self.FIN_DATE_LIST)
        if any((price_df is None, info_df is None, daily_df is None, fin_df is None)):
            print('get base em stock info failed when computing tian collection')
            return
        temp_df_list: List[pd.DataFrame] = []
        # 遍历每个datetime
        for name, group in fin_df.groupby(by='datetime'):
            # 每个group内drop掉datetime列后将其余列都加上日期后缀（包括季报和年报的时间）
            temp_df_list.append(group.set_index('stock_id').drop(columns='datetime').rename(columns=lambda x: x+'_'+name.isoformat()))
        # 将所有这些列粘到一起
        fin_df = pd.concat(temp_df_list, axis=1)

        # 在这里列出来各个table需要用到的一些字段
        info_data_list: List[str] = ['name', 'bl_em_ind_code']
        price_data_list: List[str] = ['close']
        daily_data_list: List[str] = ['total_share', 'hold_frozen_amt_accum_ratio',
                                      'pe_ttm_deducted', 'pb_lyr_n', 'ps_ttm', 'a_holder']
        fin_data_list: List[str] = ['np_ttmrp', 'performance_express_parent_ni', 'mb_sales_cons', 'mb_sales_cons_p', 'gp_margin', 'np_margin',
                                    'inv_turn_ratio', 'ar_turn_ratio', 'expense_toor', 'roe_avg', 'roe_wa', 'eps_basic', 'eps_diluted', 'bps',
                                    'balance_statement_25', 'balance_statement_46', 'balance_statement_93', 'balance_statement_103',
                                    'balance_statement_141', 'balance_statement_140', 'income_statement_9', 'income_statement_48', 'income_statement_60',
                                    'income_statement_61', 'income_statement_85', 'income_statement_127', 'income_statement_14', 'cashflow_statement_39',
                                    'cashflow_statement_59', 'cashflow_statement_77', 'cashflow_statement_82', 'cashflow_statement_86']
        # 手工拼一下带日期的列
        real_fin_data_list: List[str] = []
        for date in self.FIN_DATE_LIST:
            for fin in fin_data_list:
                real_fin_data_list.append(fin+'_'+date)
        # 取到所有需要的数据，price_df只取end_date的close
        df = pd.concat([info_df.loc[:, info_data_list], price_df[price_df.loc[:, 'datetime'] == datetime.date.fromisoformat(end_date)].loc[:, price_data_list+['stock_id']].set_index('stock_id'),
                        daily_df.loc[:, daily_data_list], fin_df.loc[:, real_fin_data_list]], axis=1)
        # 干掉没有东财分类的行
        df = df[~df['bl_em_ind_code'].isnull()]
        # 计算总市值和总资产
        df['mv'] = df['close'] * df['total_share']
        df['total_asset'] = df['balance_statement_25_2019-12-31'] + df['balance_statement_46_2019-12-31']

        # 计算主营业务
        df['mb'] = df.apply(self._get_mb, axis='columns')
        print(df.shape)
        # 计算行业数据（按东财行业分类以及按主营业务分类）
        industry_df = pd.DataFrame(columns=['bl_em_ind_code', 'pe_ttm_deducted', 'pb_lyr_n', 'ps_ttm'])
        mb_industry_df = pd.DataFrame(columns=['mb', 'pe_ttm_deducted', 'pb_lyr_n', 'ps_ttm'])
        for _, row in df.iterrows():
            # 按东财行业分类
            if row.bl_em_ind_code not in industry_df.loc[:, 'bl_em_ind_code'].to_list():
                industry_temp_df = df[row.bl_em_ind_code == df['bl_em_ind_code']].loc[:, ['pe_ttm_deducted', 'pb_lyr_n', 'ps_ttm',
                    'mv', 'total_asset', 'balance_statement_141_2019-12-31', 'income_statement_9_2019-12-31', 'income_statement_61_2019-12-31']]
                # 保存下来几个指标的行业中位数
                industry_df = industry_df.append({
                    'bl_em_ind_code': row.bl_em_ind_code,
                    'pe_ttm_deducted': industry_temp_df.loc[:, 'pe_ttm_deducted'].median(),
                    'pb_lyr_n': industry_temp_df.loc[:, 'pb_lyr_n'].median(),
                    'ps_ttm': industry_temp_df.loc[:, 'ps_ttm'].median(),
                }, ignore_index=True)
                # 保存下来每只股票中几个指标的行业排名分位数
                self._industry_rank[row.bl_em_ind_code] = pd.concat([
                    industry_temp_df.loc[:, 'pe_ttm_deducted'].rank(ascending=False, pct=True),
                    industry_temp_df.loc[:, 'pb_lyr_n'].rank(ascending=False, pct=True),
                    industry_temp_df.loc[:, 'ps_ttm'].rank(ascending=False, pct=True),
                    industry_temp_df.loc[:, 'mv'].rank(pct=True),
                    industry_temp_df.loc[:, 'total_asset'].rank(pct=True),
                    industry_temp_df.loc[:, 'balance_statement_141_2019-12-31'].rank(pct=True),
                    industry_temp_df.loc[:, 'income_statement_9_2019-12-31'].rank(pct=True),
                    industry_temp_df.loc[:, 'income_statement_61_2019-12-31'].rank(pct=True),
                ], axis='columns')
            # 按主营业务分类
            mb = self._get_mb(row)
            if mb and mb not in mb_industry_df.loc[:, 'mb'].to_list():
                mb_industry_temp_df = df[mb == df['mb']].loc[:, ['pe_ttm_deducted', 'pb_lyr_n', 'ps_ttm', 'mv', 'total_asset', 'balance_statement_141_2019-12-31', 'income_statement_9_2019-12-31', 'income_statement_61_2019-12-31']]
                mb_industry_df = mb_industry_df.append({
                    'mb': mb,
                    'pe_ttm_deducted': mb_industry_temp_df.loc[:, 'pe_ttm_deducted'].median(),
                    'pb_lyr_n': mb_industry_temp_df.loc[:, 'pb_lyr_n'].median(),
                    'ps_ttm': mb_industry_temp_df.loc[:, 'ps_ttm'].median(),
                }, ignore_index=True)
                self._mb_industry_rank[mb] = pd.concat([
                    mb_industry_temp_df.loc[:, 'pe_ttm_deducted'].rank(ascending=False, pct=True),
                    mb_industry_temp_df.loc[:, 'pb_lyr_n'].rank(ascending=False, pct=True),
                    mb_industry_temp_df.loc[:, 'ps_ttm'].rank(ascending=False, pct=True),
                    mb_industry_temp_df.loc[:, 'mv'].rank(pct=True),
                    mb_industry_temp_df.loc[:, 'total_asset'].rank(pct=True),
                    mb_industry_temp_df.loc[:, 'balance_statement_141_2019-12-31'].rank(pct=True),
                    mb_industry_temp_df.loc[:, 'income_statement_9_2019-12-31'].rank(pct=True),
                    mb_industry_temp_df.loc[:, 'income_statement_61_2019-12-31'].rank(pct=True),
                ], axis='columns')
        industry_df.set_index('bl_em_ind_code', inplace=True)

        # filter 1 读取wind四级行业分类中所有入选的股票
        stocks_in_industry = pd.read_excel('./config_tian/stocks_in_industry_2.xlsx', sheet_name='industry_lv4')
        stocks: List[str] = []
        for _, value in stocks_in_industry.loc[:, 'stocks'].items():
            for stock in value.split(';'):
                one = stock.split(',')[0]
                if one in df.index:
                    stocks.append(one)

        df = df.loc[stocks, :]
        print(f'num in stock pool: {len(df)}')
        # filter 2 按总市值过滤
        df = df[(df['mv'] >= 30*self.YI) & (df['mv'] <= 500*self.YI)]
        print(f'num after filtered by mv: {len(df)}')
        # filter 3 按控股股东质押率过滤
        df = df[df['hold_frozen_amt_accum_ratio'].isnull() | df['hold_frozen_amt_accum_ratio'] < 80]
        print(f'num after filtered by hold_frozen_amt_accum_ratio: {len(df)}')
        '''
        # filter 4 按最近三年净利润过滤
        df = df[((df['np_ttmrp_2019-12-31'].isnull()) & (df['performance_express_parent_ni_2019-12-31'] > 0) & (df['np_ttmrp_2018-12-31'] > 0) & (df['np_ttmrp_2017-12-31'] > 0)) |
                ((df['np_ttmrp_2019-12-31'] > 0) & (df['np_ttmrp_2018-12-31'] > 0) & (df['np_ttmrp_2017-12-31'] > 0))]
        print(f'num after filtered by np_ttmrp: {len(df)}')
        '''

        # scoring
        scored_df = pd.DataFrame()
        scored_df['name'] = df.loc[:, 'name']
        # 估值维度打分
        scored_df['scored_by_valuation'] = df.loc[:, ['bl_em_ind_code', 'pe_ttm_deducted', 'pb_lyr_n', 'ps_ttm']].apply(partial(self._scored_by_valuation, industry_df), axis='columns')
        # 控股股东质押率维度打分
        scored_df['scored_by_pledge_ratio'] = df.loc[:, ['hold_frozen_amt_accum_ratio']].apply(self._scored_by_pledge_ratio, axis='columns')
        # 股价维度打分
        scored_df['scored_by_price'] = df.loc[:, ['close']].apply(partial(self._scored_by_price, price_df), axis='columns')
        # 行业地位维度打分
        scored_df['scored_by_industry'] = df.loc[:, ['bl_em_ind_code', 'mb']].apply(self._scored_by_industry, axis='columns')
        # 总分扩展到百分制
        scored_df['score'] = scored_df.loc[:, ['scored_by_valuation', 'scored_by_pledge_ratio', 'scored_by_price', 'scored_by_industry']].apply(lambda x: x.sum()*100/35, axis='columns')
        scored_df['score(partial)'] = scored_df.loc[:, ['scored_by_valuation', 'scored_by_pledge_ratio', 'scored_by_price', 'scored_by_industry']].apply(lambda x: x.sum(), axis='columns')

        # 读取benchmark
        benchmark_df = pd.read_excel('./config_tian/tf_stock_pool_2th.xlsx', sheet_name='评分表', index_col=1)
        scored_df = scored_df[scored_df['name'].isin(benchmark_df['股票简称'])]
        for row in benchmark_df[['股票简称', '行业地位', '估值', '实控人', '历史股价', '总得分']].itertuples():
            scored_df.at[scored_df['name'] == row.股票简称, 'benchmark_score_by_industry'] = row.行业地位
            scored_df.at[scored_df['name'] == row.股票简称, 'benchmark_score_by_valuation'] = row.估值
            scored_df.at[scored_df['name'] == row.股票简称, 'benchmark_score_by_pledge_ratio'] = row.实控人
            scored_df.at[scored_df['name'] == row.股票简称, 'benchmark_score_by_price'] = row.历史股价
            scored_df.at[scored_df['name'] == row.股票简称, 'benchmark_score(partial)'] = row.行业地位 + row.估值 + row.实控人 + row.历史股价
            scored_df.at[scored_df['name'] == row.股票简称, 'benchmark_score'] = row.总得分
        '''
        for c in ['benchmark_score_by_industry', 'benchmark_score_by_valuation', 'benchmark_score_by_pledge_ratio', 'benchmark_score_by_price', 'benchmark_score']:
            scored_df[c] = scored_df[c].astype('int32')
        '''

        scored_df['score(with tf)'] = scored_df['score(partial)'] + scored_df['benchmark_score'] - scored_df['benchmark_score(partial)']

        # 计算总分的几个绝对误差
        scored_df['score_ae'] = scored_df['score'].sub(scored_df['benchmark_score']).abs()
        scored_df['score(with tf)_ae'] = scored_df['score(with tf)'].sub(scored_df['benchmark_score']).abs()
        scored_df['score(partial)_ae'] = scored_df['score(partial)'].sub(scored_df['benchmark_score(partial)']).abs()

        # 按几个决定误差由高到低排序
        scored_df.sort_values(by=['score(with tf)_ae', 'score(partial)_ae', 'score_ae'], ascending=False, inplace=True)
        output_list = ['name', 'scored_by_industry', 'benchmark_score_by_industry', 'scored_by_valuation', 'benchmark_score_by_valuation',
                       'scored_by_pledge_ratio', 'benchmark_score_by_pledge_ratio', 'scored_by_price', 'benchmark_score_by_price',
                       'score', 'score(with tf)', 'benchmark_score', 'score_ae', 'score(with tf)_ae', 'score(partial)', 'benchmark_score(partial)', 'score(partial)_ae']
        print(scored_df[output_list])
        scored_df[output_list].to_csv('./score.csv', float_format='%.2f')
        # df.to_sql(TianData.__tablename__, ViewDatabaseConnector().get_engine(), index=False, if_exists='append')


if __name__ == '__main__':
    TianCollectionProcessor().compute_collection('2020-02-28')




import pandas as pd
import numpy as np
import datetime
import traceback
from functools import partial
from sklearn.metrics import r2_score
import concurrent.futures
from multiprocessing import Pool
from ....data.struct import AssetTimeSpan
from ...manager.manager_fund import FundDataManager
from ...manager.score import FundScoreManager
from ...wrapper.mysql import DerivedDatabaseConnector
from ...view.derived_models import *
from .derived_data_helper import DerivedDataHelper
from ...api.derived import DerivedDataApi


class FundIndicatorProcessor(object):

    TRADING_DAYS_PER_YEAR = 242
    NATURAL_DAYS_PER_YEAR = 365
    SHORT_TIME_SPAN = TRADING_DAYS_PER_YEAR
    LONG_TIME_SPAN = TRADING_DAYS_PER_YEAR * 3
    LONG_TERM_INDEX = ['hs300', 'csi500', 'mmf']
    MIN_TIME_SPAN = int(TRADING_DAYS_PER_YEAR / 4)#为了延长基金回测范围，评分最低年限3个月
    RISK_FEE_RATE = 0.025
    RISK_FEE_RATE_PER_DAY = RISK_FEE_RATE / TRADING_DAYS_PER_YEAR

    def __init__(self, data_helper):
        self._data_helper = data_helper

    def init(self, start_date='20040101', end_date='20200630', dm=None):    
        if dm is None:
            self._dm = FundDataManager(start_time=start_date, end_time=end_date, score_manager=FundScoreManager())
            self._dm.init(score_pre_calc=False)
        else:
            self._dm = dm
        self.start_date = self._dm.start_date
        self.end_date = self._dm.end_date
        self.fund_nav = self._dm.dts.fund_nav.copy()
        self.fund_info = self._dm.dts.fund_info.copy()
        self.index_price = self._dm.dts.index_price.copy()
        
        # 指数日线价格和基金净值 时间轴对齐 ，避免join后因为日期不存在出现价格空值_rolling_alpha_beta_time_ret_r2_base
        # 指数数据历史上出现非交易日有数据，join后，基金净值出现空值，长周期时间序列算指标出空值 
        # 避免对日收益fillna(0)， 造成track error 计算不符合实际
        # 计算逻辑空值填充情况： 1. 只对基金费率空值fillna(0) 
        #                    2. manager fund 里对日线价格ffill
        #                    3. 指数数据按照基金数据重做index后ffill(有基金数据对日子,没有指数数据，比如季报出在季度末，非交易日有基金净值)
        #                    4. 有对超过基金终止日净值赋空值
        self.trading_days = self._dm.dts.trading_days.datetime
        end_date = min(self.fund_nav.index[-1], self.index_price.index[-1])
        self.fund_nav = self.fund_nav.reindex(self.trading_days).loc[:end_date]
        self.index_price = self.index_price.reindex(self.trading_days).loc[:end_date]
        fund_nav_index = self.fund_nav.index
        self.index_price = self.index_price.reindex(fund_nav_index).fillna(method = 'ffill')
        l1 = self.fund_nav.index.to_list().sort()
        l2 = self.index_price.index.to_list().sort()
        assert l1 == l2, f'date index of index price and fund nav are not identical, l1 {l1} lens {len(l1)}, l2 {l2} lens {len(l2)}'

        fund_to_enddate_dict = self.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        self.fund_to_index_dict = self.fund_info[['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']
        # 超过基金终止日的基金净值赋空
        for fund_id in self.fund_nav.columns:
            fund_end_date = fund_to_enddate_dict[fund_id]
            if self.end_date > fund_end_date:
                self.fund_nav.loc[fund_end_date:,fund_id] = np.nan
        self.fund_ret = np.log(self.fund_nav / self.fund_nav.shift(1))
        self.fund_ret = self.fund_ret.stack().reset_index(level=[0,1]).rename(columns={0:'ret'})
        self.fund_ret = self.fund_ret.pivot_table(index='datetime',columns='fund_id',values='ret')        
        # self.fund_ret['index_id'] = self.fund_ret.fund_id.apply(lambda x: self.fund_to_index_dict[x])
        # self.fund_ret = self.fund_ret.pivot_table(index = ['index_id','datetime'],columns='fund_id',values='ret')
        self.index_ret = np.log(self.index_price / self.index_price.shift(1))
        self.fund_to_index_dict = {fund_id:index_id for fund_id,index_id in self.fund_to_index_dict.items() if fund_id in self.fund_ret.columns}
        self.index_list = list(AssetTimeSpan.__dataclass_fields__.keys())
        self.index_fund = { index_id : [fund_idx for fund_idx, index_idx in self.fund_to_index_dict.items() if index_idx == index_id] for index_id in self.index_list}
        self.index_fund['conv_bond'] = self._dm.dts.fund_conv_list
        self.start_date_dic = self.fund_info[['fund_id','start_date']].set_index('fund_id').to_dict()['start_date']

    def get_time_range(self, index_id):
        if index_id in self.LONG_TERM_INDEX:
            return self.LONG_TIME_SPAN
        else:
            return self.SHORT_TIME_SPAN

    ''' 
    尝试过几个版本的回归函数， 计算结果一致
    1. 最早使用的是pyfinance.PandasRollingOLS  可以对一只基金直接返回beta, alpha, 可以设置滚动周期, 运算比较快， 缺点是无法设置min_periods，最小滚动周期
    2. pyfinance.PandasRollingOLS 本质是使用statsmodels.regression.rolling.RollingOLS, 需要手动增加回归常数项，函数下存在最小窗口设置，但是使用无效，
       而且window实际的含义和pandas.rolling.window有区别，导致计算有效值和nan的边界时间不一致
    3. np.polyfit 一元回归，自带回归常数项，可以设置回归最高幂
    '''

    def _rolling_alpha_beta_time_ret_r2(self, x, res, df, is_only_alpha_beta=False):
        # 回归相关的都在这里
        df_i = df.loc[x[0]:x[-1],]
        return self._rolling_alpha_beta_time_ret_r2_base(res,df_i,is_only_alpha_beta)

    def _rolling_alpha_beta_time_ret_r2_base(self, res, df_i, is_only_alpha_beta=False):
        if sum(df_i.fund) == 0:
            res.append({'alpha':np.Inf,'beta':np.Inf,'time_ret':np.Inf})
            return 1
        else:
            ploy_res = np.polyfit(y=df_i.fund, x=df_i.benchmark,deg=1)
            p = np.poly1d(ploy_res)
            r2 = r2_score(df_i.fund, p(df_i.benchmark))
            beta = ploy_res[0]
            alpha = ploy_res[1] * self.TRADING_DAYS_PER_YEAR
            if is_only_alpha_beta:
                res.append({'alpha':alpha,'beta':beta})
                return 1
            day_len = df_i.shape[0]
            bar_num = int( day_len / self.MIN_TIME_SPAN)

            _res = []
            for i in range(bar_num):
                start_i = - (i + 1) * self.MIN_TIME_SPAN
                end_i = - i * self.MIN_TIME_SPAN
                if end_i == 0:
                    dftmp = df_i.iloc[start_i:]
                else:
                    dftmp = df_i.iloc[start_i:end_i]
                _ploy_res = np.polyfit(y=dftmp.fund, x=dftmp.benchmark,deg=1)
                
                _res.append({'beta_i_no_whole_beta': _ploy_res[0] - beta,
                            'bench_r_no_risk': dftmp.benchmark.sum() - self.RISK_FEE_RATE_PER_DAY * day_len })
            time_ret = np.sum([ _['beta_i_no_whole_beta'] * _['bench_r_no_risk'] for _ in _res])
            res.append({'alpha': alpha, 
                        'beta': beta,
                        'time_ret':time_ret,
                        'r_square':r2})
            return 1

    def _rolling_mdd(self, x):
        x = pd.Series(x)
        return 1 - (x / x.cummax()).min()

    def _rolling_annual_ret(self, x):
        x = pd.Series(x).dropna()
        year = x.shape[0] / self.TRADING_DAYS_PER_YEAR
        return np.exp(np.log(x.values[-1]/x.values[0])/year) - 1
 
    def _process_fund_indicator_update(self, fund_id, time_range, fund_ret, index_id, index_ret):
        df = fund_ret[[fund_id]].join(index_ret).dropna()
        if df.empty:
            return
        df = df.rename(columns={index_id:'benchmark',fund_id:'fund'}).reset_index()
        last_day = df.datetime.values[-1]
        year_length = (last_day-self.start_date_dic[fund_id]).days / self.NATURAL_DAYS_PER_YEAR
        df = df.iloc[-time_range:].dropna()
        if df.shape[0] < self.MIN_TIME_SPAN:
            return None
        res = []
        self._rolling_alpha_beta_time_ret_r2_base(res, df)
        fund_indicator_part1 = res[0]
        if fund_indicator_part1['alpha'] in [np.Inf, -np.Inf]:
            return None
        df_hs300 = self.fund_ret[[fund_id]].join(self.index_ret['hs300']).rename(columns={'hs300':'benchmark',fund_id:'fund'}).reset_index().iloc[-time_range:].dropna()
        df_bond = self.fund_ret[[fund_id]].join(self.index_ret['national_debt']).rename(columns={'national_debt':'benchmark',fund_id:'fund'}).reset_index().iloc[-time_range:].dropna()
        res = []
        self._rolling_alpha_beta_time_ret_r2_base(res, df_hs300, is_only_alpha_beta=True)
        fund_indicator_hs300 = res[0]
        res = []
        self._rolling_alpha_beta_time_ret_r2_base(res, df_bond, is_only_alpha_beta=True)
        fund_indicator_bond = res[0]
        track_err = (df.fund - df.benchmark).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        timespan = int(time_range / self.TRADING_DAYS_PER_YEAR)
        fee_rate = self.fund_fee[fund_id]
        info_ratio = fund_indicator_part1['alpha'] / track_err
        mean_ret = df.fund.mean()
        mean_ret_no_free_ret = mean_ret - self.RISK_FEE_RATE_PER_DAY
        treynor = mean_ret_no_free_ret * time_range / fund_indicator_part1['beta']
        if self.fund_nav[[fund_id]][-time_range:].dropna().shape[0] < self.MIN_TIME_SPAN:
            return None
        mdd = self.fund_nav[[fund_id]][-time_range:].apply(self._rolling_mdd, raw=True).values[0]
        down_risk = np.abs(np.minimum(df['fund'] - self.RISK_FEE_RATE_PER_DAY, 0)).mean()* np.sqrt(self.TRADING_DAYS_PER_YEAR)
        ret_over_period = self.fund_nav[fund_id][-1] / self.fund_nav[fund_id].fillna(method='bfill')[-time_range-1] - 1
        annual_avg_daily_ret = df['fund'].mean() * self.TRADING_DAYS_PER_YEAR 
        annual_ret = self.fund_nav[[fund_id]][-time_range:].apply(self._rolling_annual_ret, raw=True)[0]
        annual_vol = df['fund'].std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        vol_benchmark = df['benchmark'].std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        mean_ret_benchmark = df['benchmark'].mean()
        m_square = vol_benchmark / annual_vol * mean_ret_no_free_ret + self.RISK_FEE_RATE_PER_DAY - mean_ret_benchmark
        var = np.minimum(df['fund'].quantile(0.05), 0) * -1
        sharpe = (annual_ret - self.RISK_FEE_RATE) / annual_vol
        dic = {
            'fund_id':fund_id,
            'datetime':last_day,
            'beta':fund_indicator_part1['beta'],
            'alpha':fund_indicator_part1['alpha'],
            'time_ret':fund_indicator_part1['time_ret'],
            'r_square':fund_indicator_part1['r_square'],
            'track_err':track_err,
            'timespan':timespan,
            'fee_rate':fee_rate,
            'info_ratio':info_ratio,
            'treynor':treynor,
            'mdd':mdd,
            'down_risk':down_risk,
            'ret_over_period':ret_over_period,
            'annual_avg_daily_ret':annual_avg_daily_ret,
            'annual_ret':annual_ret,
            'annual_vol':annual_vol,
            'm_square':m_square,
            'var':var,
            'sharpe':sharpe,
            'year_length':year_length,
            'alpha_bond':fund_indicator_bond['alpha'],
            'beta_bond':fund_indicator_bond['beta'],
            'alpha_hs300':fund_indicator_hs300['alpha'],
            'beta_hs300':fund_indicator_hs300['beta'],}
        return dic

    def _process_fund_indicator_one(self, time_range, fund_ret, index_id, index_ret, fund_id):
        try:
            df = fund_ret[[fund_id]].join(index_ret).dropna()
            df = df.rename(columns={index_id:'benchmark',fund_id:'fund'}).reset_index()
            df['year_length'] = df['datetime'].map(lambda x: (x-self.start_date_dic[fund_id]).days / self.NATURAL_DAYS_PER_YEAR)
            res = []
            pd.Series(df.index).rolling(
                window=time_range,min_periods=self.MIN_TIME_SPAN).apply(
                    partial(self._rolling_alpha_beta_time_ret_r2, res=res, df=df), raw=True)
            df = df.set_index('datetime')
            df = df.join(pd.DataFrame(res,index=df.index[-len(res):]))
            if 'alpha' not in df.columns:
                return None
            df_hs300 = self.fund_ret[[fund_id]].join(self.index_ret['hs300']).rename(columns={'hs300':'benchmark',fund_id:'fund'}).reset_index().dropna()
            df_bond = self.fund_ret[[fund_id]].join(self.index_ret['national_debt']).rename(columns={'national_debt':'benchmark',fund_id:'fund'}).reset_index().dropna()
            res_hs300 = []
            pd.Series(df_hs300.index).rolling(
                window=time_range,min_periods=self.MIN_TIME_SPAN).apply(
                    partial(self._rolling_alpha_beta_time_ret_r2, res=res_hs300, df=df_hs300, is_only_alpha_beta=True), raw=True)
            res_hs300 = pd.DataFrame(res_hs300, index=df_hs300.datetime[-len(res_hs300):]).rename(columns={'alpha':'alpha_hs300','beta':'beta_hs300'})
            res_bond = []
            pd.Series(df_bond.index).rolling(
                window=time_range,min_periods=self.MIN_TIME_SPAN).apply(
                    partial(self._rolling_alpha_beta_time_ret_r2, res=res_bond, df=df_bond, is_only_alpha_beta=True), raw=True)
            res_bond = pd.DataFrame(res_bond, index=df_bond.datetime[-len(res_bond):]).rename(columns={'alpha':'alpha_bond','beta':'beta_bond'})
            df = df.join(res_hs300).join(res_bond)
            df['track_err'] = (df.fund - df.benchmark).rolling(window=time_range, min_periods=self.MIN_TIME_SPAN).std(ddof=1)* np.sqrt(self.TRADING_DAYS_PER_YEAR)
            df['fund_id'] = fund_id
            df['timespan'] = int(time_range / self.TRADING_DAYS_PER_YEAR)
            df['fee_rate'] = self.fund_fee[fund_id]
            df['info_ratio'] = df.alpha / df.track_err
            df['mean_ret'] = df[['fund']].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).mean()
            df['mean_ret_no_free_ret'] = df['mean_ret'] - self.RISK_FEE_RATE_PER_DAY
            df['treynor'] = df['mean_ret_no_free_ret'] * time_range / df.beta
            df['mdd'] = self.fund_nav[[fund_id]].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).apply(self._rolling_mdd, raw=True)
            df['down_risk'] = np.abs(np.minimum(df['fund'] - self.RISK_FEE_RATE_PER_DAY, 0))
            df['down_risk'] = df['down_risk'].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).mean()* np.sqrt(self.TRADING_DAYS_PER_YEAR)
            df['ret_over_period'] = self.fund_nav[[fund_id]] / self.fund_nav[[fund_id]].fillna(method='bfill').shift(time_range) - 1
            df['annual_avg_daily_ret'] = df['fund'].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).mean() * self.TRADING_DAYS_PER_YEAR 
            df['annual_ret'] = self.fund_nav[[fund_id]].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).apply(self._rolling_annual_ret, raw=True)
            df['annual_vol'] = df['fund'].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
            df['vol_benchmark'] = df['benchmark'].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
            df['mean_ret_benchmark'] = df[['benchmark']].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).mean()
            df['m_square'] = df['vol_benchmark'] / df['annual_vol'] * df['mean_ret_no_free_ret'] + self.RISK_FEE_RATE_PER_DAY - df['mean_ret_benchmark']
            df['var'] = df['fund'].rolling(window=time_range,min_periods=self.MIN_TIME_SPAN).quantile(0.05)
            df['var'] = np.minimum(df['var'], 0) * -1
            df['sharpe'] = (df.annual_ret - self.RISK_FEE_RATE) / df.annual_vol
            df = df.drop(['fund','benchmark','vol_benchmark','mean_ret_no_free_ret','mean_ret_benchmark','mean_ret'], axis=1)
            # build history insert
            df = df.replace({-np.Inf:None,np.Inf:None}).dropna(subset=['beta', 'alpha','track_err']).reset_index()
            #self._data_helper._upload_derived(df, FundIndicator.__table__.name)
        except:
            print(f'fund_id {fund_id} calc indicator error')
        return df

    def calculate_history(self, index_id_list:str=None, cpu_num:int=2):
        self.fund_fee = self.fund_info[['fund_id','manage_fee','trustee_fee']].set_index('fund_id').fillna(0).sum(axis = 1)
        result = []
        index_list = self.index_list if index_id_list is None else index_id_list
        fund_index_dict = {}
        for index_id in index_list:
            time_range = self.get_time_range(index_id)
            if index_id in ['active']:
                continue
            if index_id in ['conv_bond']:
                fund_list = self.index_fund[index_id]
                fund_list = [i for i in fund_list if i not in fund_index_dict]
                fund_list = [i for i in fund_list if i in self.fund_ret.columns]
            else:
                fund_list_1 = self.index_fund[index_id]
                fund_list_2 = self._dm.dts.active_fund_info[self._dm.dts.active_fund_info['index_id'] == index_id].fund_id.tolist()
                fund_list = list(set(fund_list_1).union(set(fund_list_2)).intersection(self._dm.dts.all_fund_list))
            fund_ret = self.fund_ret[fund_list].copy()
            index_ret = self.index_ret[[index_id]]
            p = Pool()
            res = [i for i in p.imap_unordered(partial(self._process_fund_indicator_one, time_range, fund_ret, index_id, index_ret), fund_list, 256) if i is not None]
            p.close()
            p.join()
            result.extend(res)    
            print(f'index {index_id} fund number : {len(fund_list)} finish' )
            for fund_id in fund_list:
                fund_index_dict[fund_id] = index_id
        result = [ _ for _ in result if not _ is None]
        self.result = pd.concat(result, axis=0, sort=True)

    def calculate_update(self, dt):
        self.fund_fee = self.fund_info[['fund_id','manage_fee','trustee_fee']].set_index('fund_id').fillna(0).sum(axis = 1)
        result = []
        index_list = self.index_list
        fund_index_dict = {}
        for index_id in index_list:
            time_range = self.get_time_range(index_id)
            if index_id in ['active']:
                continue
            if index_id in ['conv_bond']:
                fund_list = self.index_fund[index_id]
                fund_list = [i for i in fund_list if i not in fund_index_dict]
                fund_list = [i for i in fund_list if i in self.fund_ret.columns]
            else:
                fund_list_1 = self.index_fund[index_id]
                fund_list_2 = self._dm.dts.active_fund_info[self._dm.dts.active_fund_info['index_id'] == index_id].fund_id.tolist()
                fund_list = list(set(fund_list_1).union(set(fund_list_2)).intersection(self._dm.dts.all_fund_list))
            fund_list = list(self.fund_ret.columns.intersection(fund_list))
            fund_ret = self.fund_ret[fund_list].copy()
            index_ret = self.index_ret[[index_id]]
            for fund_id in fund_list:
                result.append(self._process_fund_indicator_update(fund_id, time_range, fund_ret.loc[:dt], index_id, index_ret.loc[:dt]))
            print(f'index {index_id} finish fund number fund_list {len(fund_list)} fund_list_1 {len(fund_list_1)} fund_list_2 {len(fund_list_2)}')
            for fund_id in fund_list:
                fund_index_dict[fund_id] = index_id
        result = [ _ for _ in result if not _ is None]
        self.result = pd.DataFrame(result)
        self.result = self.result.replace({-np.Inf:None,np.Inf:None}).dropna(subset=['beta', 'alpha','track_err'])

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            start_date_dt = datetime.datetime.strptime(start_date, '%Y%m%d').date()
            end_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()

            start_date = start_date_dt - datetime.timedelta(days = 1150) #3年历史保险起见，多取几天 3*365=1095 
            start_date = datetime.datetime.strftime(start_date, '%Y%m%d')

            self.init(start_date=start_date, end_date=end_date)

            self._update_begin_date = self._dm.dts.trading_days[self._dm.dts.trading_days.datetime <= self.end_date] 
            update_dt_list = self._update_begin_date.datetime.values[-3:].tolist()
            self.fac_result = []
            for dt in update_dt_list:
                self.calculate_update(dt)
                self.fac_result.append(self.result)
            df = pd.concat(self.fac_result)
            df = df[df.datetime.isin(update_dt_list)]
            df = df.drop_duplicates(subset=['datetime','fund_id'])
            for date in df.datetime.unique():
                DerivedDataApi().delete_fund_indicator_with_dt(date, df[df.datetime == date].fund_id.to_list())
            self._data_helper._upload_derived(df, FundIndicator.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_indicator')

        return failed_tasks


if __name__ == '__main__':
    fip = FundIndicatorProcessor(DerivedDataHelper())
    fip.process('20210602', '20210602')



import numpy as np
from contextlib import contextmanager

from sqlalchemy import create_engine, event
from sqlalchemy.orm import scoped_session, sessionmaker

from ...util.singleton import Singleton
from ...util.config import SurfingConfigurator

class RawDatabaseConnector(metaclass=Singleton):

    def __init__(self,db_tag=None):
        db_conn_str = SurfingConfigurator().get_db_settings(db_name='raw').to_conn_str()
        self.engine = create_engine(db_conn_str, pool_size=20, max_overflow=20, echo_pool=True)
        self.smaker = sessionmaker(bind=self.engine)
        self.scoped_smaker = scoped_session(self.smaker)

    @contextmanager
    def managed_session(self):
        session = self.scoped_smaker()
        try:
            yield session
        finally:
            session.close()

    def get_engine(self):
        return self.engine


class BasicDatabaseConnector(metaclass=Singleton):

    def __init__(self,db_tag=None):
        db_conn_str = SurfingConfigurator().get_db_settings(db_name='basic').to_conn_str()
        self.engine = create_engine(db_conn_str, pool_size=20, max_overflow=20, echo_pool=True)
        self.smaker = sessionmaker(bind=self.engine)
        self.scoped_smaker = scoped_session(self.smaker)

    @contextmanager
    def managed_session(self):
        session = self.scoped_smaker()
        try:
            yield session
        finally:
            session.close()

    def get_engine(self):
        return self.engine


class DerivedDatabaseConnector(metaclass=Singleton):

    def __init__(self,db_tag=None):
        db_conn_str = SurfingConfigurator().get_db_settings(db_name='derived').to_conn_str()
        self.engine = create_engine(db_conn_str, pool_size=20, max_overflow=20, echo_pool=True)
        self.smaker = sessionmaker(bind=self.engine)
        self.scoped_smaker = scoped_session(self.smaker)

    @contextmanager
    def managed_session(self):
        session = self.scoped_smaker()
        try:
            yield session
        finally:
            session.close()

    def get_engine(self):
        return self.engine


class ViewDatabaseConnector(metaclass=Singleton):

    def __init__(self):
        self.db_conn_str = SurfingConfigurator().get_db_settings(db_name='view').to_conn_str()
        self._connect()

    def _connect(self):
        self.engine = create_engine(
            self.db_conn_str,
            pool_size=20,
            max_overflow=20,
            echo_pool=True,
            pool_recycle=3600,
            pool_pre_ping=True,
        )
        self.session_maker = sessionmaker(bind=self.engine)
        self.scoped_maker = scoped_session(self.session_maker)

    @contextmanager
    def managed_session(self):
        if not self.engine:
            self._connect()
        session = self.scoped_maker()
        try:
            yield session
        finally:
            session.close()

    def get_engine(self):
        return self.engine

class TestDatabaseConnector(metaclass=Singleton):

    def __init__(self,db_tag=None):
        db_conn_str = SurfingConfigurator().get_db_settings(db_name='test').to_conn_str()
        self.engine = create_engine(db_conn_str, pool_size=20, max_overflow=20, echo_pool=True)
        self.smaker = sessionmaker(bind=self.engine)
        self.scoped_smaker = scoped_session(self.smaker)

    @contextmanager
    def managed_session(self):
        session = self.scoped_smaker()
        try:
            yield session
        finally:
            session.close()

    def get_engine(self):
        return self.engine

class TLDatabaseConnector(metaclass=Singleton):

    def __init__(self,db_tag=None):
        db_conn_str = SurfingConfigurator().get_db_settings(db_name='TL').to_conn_str()
        self.engine = create_engine(db_conn_str, pool_size=20, max_overflow=20, echo_pool=True)
        self.smaker = sessionmaker(bind=self.engine)
        self.scoped_smaker = scoped_session(self.smaker)
        event.listen(self.engine, "before_cursor_execute", self.add_own_encoders)

    def add_own_encoders(self, conn, cursor, query, *args):
        cursor.connection.encoders[np.int64] = lambda value, encoders: int(value)

    @contextmanager
    def managed_session(self):
        session = self.scoped_smaker()
        try:
            yield session
        finally:
            session.close()

    def get_engine(self):
        return self.engine

class RawFutureMinDatabaseConnector(metaclass=Singleton):

    def __init__(self,db_tag=None):
        db_conn_str = SurfingConfigurator().get_db_settings(db_name='raw_future_min').to_conn_str()
        self.engine = create_engine(db_conn_str, pool_size=20, max_overflow=20, echo_pool=True)
        self.smaker = sessionmaker(bind=self.engine)
        self.scoped_smaker = scoped_session(self.smaker)
        event.listen(self.engine, "before_cursor_execute", self.add_own_encoders)

    def add_own_encoders(self, conn, cursor, query, *args):
        cursor.connection.encoders[np.int64] = lambda value, encoders: int(value)

    @contextmanager
    def managed_session(self):
        session = self.scoped_smaker()
        try:
            yield session
        finally:
            session.close()

    def get_engine(self):
        return self.engine


from sqlalchemy import CHAR, Column, Integer, TEXT, BOOLEAN, text, Index, Enum
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.dialects.mysql import DOUBLE, DATE, TINYINT, DATETIME, MEDIUMTEXT, YEAR, SMALLINT, BIGINT, DECIMAL, VARCHAR, TIMESTAMP

from ...constant import SectorType, IndexPriceSource, CodeFeeMode, IndClassType

class Base:
    pass
    #_update_time = Column('_update_time', DATETIME, nullable=False, server_default=text('CURRENT_TIMESTAMP ON UPDATE CURRENT_TIMESTAMP'))  # 更新时间
# make this column at the end of every derived table
#Base._update_time._creation_order = 9999
Base = declarative_base(cls=Base)


class PfundInstScaleAmac(Base):
    '''私募基金管理人规模'''

    __tablename__ = 'pfund_inst_scale_amac'
    
    ID = Column(BIGINT, primary_key=True) # 信息编码
    PARTY_ID = Column(BIGINT) # 机构内部编码
    PARTY_FULL_NAME = Column(CHAR(300)) # 机构名称 
    MAIN_FUND_TYPE = Column(CHAR(200)) # 管理基金主要类别 
    SCALE = Column(CHAR(50)) # 管理规模 
    INST_ID = Column(CHAR(60)) # 源数据id 
    NUM = Column(Integer) # 变更次数 
    IS_NEW = Column(Integer) # 是否最新
    FETCH_DATE = Column(DATE) # 变动时间 
    UPDATE_TIME = Column(DATETIME) # 更新时间 
    TMSTAMP = Column(BIGINT) # 后台系统修改数据时间戳
    HERMESFIRSTTIME = Column(TIMESTAMP) # 记录数据首次写入时间
    HERMESUPTIME = Column(TIMESTAMP) # 记录数据更新时间
    

class PfundManagerInfo(Base):
    '''私募基金经理基本信息表'''

    __tablename__ = 'pfund_manager_info'
    
    ID = Column(BIGINT, primary_key=True) # 信息编码
    PERSON_ID = Column(CHAR(60)) # 人员ID
    NAME = Column(CHAR(60)) # 姓名 
    GENDER = Column(CHAR(60)) # 性别 
    BIRTHDAY = Column(CHAR(60)) # 生日 
    NATIONALITY_CD = Column(CHAR(60)) # 国籍 
    CERTIFICATE = Column(CHAR(300)) # 证书 
    PHOTO = Column(CHAR(200)) # 照片
    EDUCATION = Column(Integer) # 最高学历 
    GRAD_UNIV = Column(CHAR(60)) # 毕业院校 
    MAJOR = Column(CHAR(60)) # 所学专业
    BACKGROUND_DESC = Column(TEXT) # 简介
    BACKGROUND = Column(CHAR(60)) # 背景
    MF_TO_PF = Column(TINYINT) # 是否公转私/私转公
    YEAR = Column(DECIMAL(4,2)) # 投资年限
    AWARDS_NAME = Column(CHAR(2000)) # 所获荣誉
    DEPARTMENT = Column(CHAR(60)) # 所在部门
    POSITION = Column(CHAR(500)) # 职务
    IS_QUAL = Column(Integer) # 是否具备基金从业资格
    NUM_DURA = Column(BIGINT) # 当前管理的基金数量
    DURA_SEC_ID = Column(MEDIUMTEXT) # 当前管理的基金ID
    DURA_SEC_NAME = Column(MEDIUMTEXT) # 当前管理的基金
    PARTY_ID = Column(BIGINT) # 任职机构ID
    PARTY_FULL_NAME = Column(CHAR(300)) # 任职机构
    REP_SEC_ID = Column(BIGINT) # 代表产品ID
    REP_SEC_NAME = Column(CHAR(300)) # 代表产品
    RETURN_MAN_A = Column(DECIMAL(18,10)) # 任职收益
    REMARK = Column(CHAR(300)) # 备注
    UPDATE_TIME = Column(DATETIME) # 更新时间
    TMSTAMP = Column(BIGINT) # 后台系统修改数据时间戳
    HERMESFIRSTTIME = Column(TIMESTAMP) # 记录数据首次写入时间
    HERMESUPTIME = Column(TIMESTAMP) # 记录数据更新时间
    
    
class PfundNav(Base):
    '''私募净值'''

    __tablename__ = 'pfund_nav'
    
    ID = Column(BIGINT, primary_key=True) # 自增ID
    SECURITY_ID = Column(BIGINT) # 证券内部编码
    PUBLISH_DATE = Column(DATE) # 数据发布日期 
    END_DATE = Column(DATE) # 净值日期 
    NAV = Column(DECIMAL(19,4)) # 单位净值 
    ACCUM_NAV = Column(DECIMAL(19,4)) # 累计净值 
    UPDATE_TIME = Column(DATETIME) # 更新时间 
    TMSTAMP = Column(BIGINT) # 后台系统修改数据时间戳
    ADJ_NAV = Column(DECIMAL(19,4)) # 调整净值 
    NAV_UNIT = Column(DECIMAL(19,4)) # 净值单位 
    RETURN_RATE = Column(DECIMAL(19,10)) # 净值收益 
    ADJ_RETURN_RATE = Column(DECIMAL(19,4)) # 调整净值收益 
    HERMESFIRSTTIME = Column(TIMESTAMP) # 记录数据首次写入时间
    HERMESUPTIME = Column(TIMESTAMP) # 记录数据更新时间


class SysCode(Base):
    '''参数集合'''

    __tablename__ = 'sys_code'
    
    ID = Column(BIGINT, primary_key=True) # 自增ID
    CODE_TYPE_ID = Column(Integer) # 参数ID
    CODE_TYPE_NAME = Column(CHAR(100)) # 参数名称 
    CODE_TYPE_NAME_EN = Column(CHAR(100)) # 参数英文名称 
    VALUE_NO = Column(SMALLINT) # 参数值序列 
    VALUE_EN_CD = Column(CHAR(50)) # 参数值字母编码 
    VALUE_NUM_CD = Column(CHAR(20)) # 参数值数字编码 
    VALUE_NAME_CN = Column(CHAR(100)) # 参数值中文名称 
    VALUE_NAME_EN = Column(CHAR(100)) # 参数值英文缩写 
    REMARK = Column(CHAR(500)) # 备注 
    TMSTAMP = Column(BIGINT) #  后台系统修改数据时间戳
    UPDATE_TIME = Column(DATETIME) # 更新时间
    HERMESFIRSTTIME = Column(TIMESTAMP) # 记录数据首次写入时间
    HERMESUPTIME = Column(TIMESTAMP) # 记录数据更新时间

class MdInstitution(Base):
    '''公司基本信息'''
        
    __tablename__ = 'md_institution'
    
    PARTY_ID = Column(Integer, primary_key=True) # 机构内部ID
    PARTY_FULL_NAME = Column(CHAR(100)) # 机构全称 
    PARTY_SHORT_NAME = Column(CHAR(100)) # 机构简称 
    PARTY_FULL_NAME_EN = Column(CHAR(200)) # 机构英文全称 
    PARTY_SHORT_NAME_EN = Column(CHAR(200)) # 机构英文简称 
    REG_DATE = Column(DATE) # 注册日期 
    REG_COUNTRY_CD = Column(CHAR(3)) # 注册国家代码 
    REG_PROVINCE = Column(CHAR(100)) # 注册省份 
    REG_CITY = Column(CHAR(100)) # 注册城市 
    REG_ADDR = Column(CHAR(200)) # 注册地址 
    REG_CAP = Column(DECIMAL(18,4)) # 注册资金 
    REG_CAP_CURR_CD = Column(CHAR(3)) # 注册资金货币代码 
    OFFICE_ADDR = Column(CHAR(200)) # 办公地址 
    EMAIL = Column(CHAR(100)) # 联系邮箱 
    WEBSITE = Column(CHAR(100)) # 机构网站 
    TEL = Column(CHAR(50)) # 联系电话 
    FAX = Column(CHAR(50)) # 联系传真 
    LEGAL_ENTITY_ID = Column(CHAR(20)) # 工商登记号 
    PRIME_OPERATING = Column(MEDIUMTEXT) # 主营业务 
    PARTY_NATURE_CD = Column(CHAR(100)) # 公司性质 
    IS_ISS_BOND = Column(CHAR(100)) # 是否发行债券 
    CLOSE_DATE = Column(DATE) # 机构终止日期 
    INST_STATUS = Column(SMALLINT) # 是否运作 
    DY_USE_FLG = Column(BIGINT) # DY使用标识 
    UPDATE_TIME = Column(DATETIME) # 更新时间 
    TMSTAMP = Column(BIGINT) #  后台系统修改数据时间戳
    PROFILE = Column(TEXT) # 机构介绍 
    BOARD_SECRY = Column(CHAR(50)) # 董事会秘书 
    LEGAL_REP = Column(CHAR(50)) # 法定代表人 
    OPA_SCOPE = Column(MEDIUMTEXT) # 经营范围 
    GEN_MANA = Column(CHAR(50)) # 总经理 
    IS_ISS_MF = Column(Integer) # 是否发行公募基金 
    HERMESFIRSTTIME = Column(TIMESTAMP) #  记录数据首次写入时间
    HERMESUPTIME = Column(TIMESTAMP) #  记录数据更新时间
    

class MdSecurity(Base):
    '''主编码及基本信息'''
        
    __tablename__ = 'md_security'
    
    SECURITY_ID = Column(Integer, primary_key=True) # 证券内部ID
    TICKER_SYMBOL = Column(CHAR(50)) # 交易代码 
    EXCHANGE_CD = Column(CHAR(4)) # 交易市场代码 
    SEC_FULL_NAME = Column(CHAR(100)) # 证券全称 
    SEC_SHORT_NAME = Column(CHAR(100)) # 证券简称 
    CN_SPELL = Column(CHAR(200)) # 简称拼音 
    SEC_FULL_NAME_EN = Column(CHAR(200)) # 证券英文全称 
    SEC_SHORT_NAME_EN = Column(CHAR(200)) # 证券英文简称 
    DYID = Column(CHAR(50)) # 证券展示代码 
    EX_COUNTRY_CD = Column(CHAR(3)) # 交易市场所属地区 
    TRANS_CURR_CD = Column(CHAR(3)) # 交易货币代码 
    ASSET_CLASS = Column(CHAR(4)) # 证券类型 
    LIST_STATUS_CD = Column(CHAR(4)) # 上市状态 
    LIST_DATE = Column(DATE) # 上市日期 
    DELIST_DATE = Column(DATE) # 摘牌日期 
    PARTY_ID = Column(BIGINT) # 机构内部ID 
    DY_USE_FLG = Column(BIGINT) # DY使用标识 
    UPDATE_TIME = Column(DATETIME) # 更新时间 
    TMSTAMP = Column(BIGINT) #  后台系统修改数据时间戳
    HERMESFIRSTTIME = Column(TIMESTAMP) #  记录数据首次写入时间
    HERMESUPTIME = Column(TIMESTAMP) #  记录数据更新时间
     
                  
class Pfund(Base):
    '''私募基本信息'''
        
    __tablename__ = 'pfund'
    
    ID = Column(BIGINT, primary_key=True) # 自增ID

    SECURITY_ID = Column(BIGINT) # 证券内部编码 
    ESTABLISH_DATE = Column(DATE) # 成立日期 
    PF_STYLE = Column(CHAR(10)) # 产品类型 
    STATUS = Column(CHAR(200)) # 运行状态 
    INVEST_STRATEGY = Column(CHAR(10)) # 投资策略 
    INVEST_STRATEGY_CHILD = Column(CHAR(200)) # 子策略 
    DURATION = Column(Integer) # 存续期限 
    OPEN_DATE_DESC = Column(CHAR(500)) # 开放日 
    INVEST_CONSULTANT = Column(CHAR(50)) # 投资顾问 
    CUSTODIAN = Column(CHAR(50)) # 托管人 
    ISSUE_PLATFORM = Column(CHAR(100)) # 受托人 
    TRADING_BROKER = Column(CHAR(50)) # 交易经纪 
    SUBSCRIPTION_START_POINT = Column(Integer) # 认购起点 
    SCALE_INITIAL = Column(BIGINT) # 初始规模 
    ISSUE_FEE = Column(DECIMAL(19,10)) # 认购费率 
    REDEEM_FEE = Column(DECIMAL(19,10)) # 赎回费率 
    MANAGEMENT_FEE = Column(DECIMAL(19,10)) # 管理费率 
    PERFORMANECE_RETURN = Column(DECIMAL(19,10)) # 业绩报酬 
    RECORD_CD = Column(CHAR(50)) # 备案编码 
    UPDATE_TIME = Column(DATETIME) # 更新时间 
    TMSTAMP = Column(BIGINT) #  后台系统修改数据时间戳
    RECORD_DATE = Column(DATE) # 备案日期 
    END_DATE = Column(DATE) # 到期日 
    NAV_FREQ = Column(CHAR(20)) # 净值发布频率 
    CLOSING_DURA_DESC = Column(CHAR(200)) # 封闭期 
    ISSUE_LOC = Column(CHAR(200)) # 发行地 
    RECORD_STATUS = Column(CHAR(20)) # 备案状态 
    MANAGE_TYPE = Column(TINYINT(3)) # 管理类型 
    MANAGER = Column(CHAR(500)) # 基金经理 
    MIN_ADD = Column(CHAR(200)) # 最低追加金额 
    STOP_LOSS_LINE = Column(DECIMAL(19,4)) # 止损线 
    WARN_LINE = Column(DECIMAL(19,4)) #  预警线
    FUTURES_BROKER = Column(CHAR(50)) #  期货经纪
    AMAC_TYPE = Column(CHAR(200)) #  协会公布类型
    APPLY_FEE = Column(DECIMAL(19,4)) # 申购费
    CUSTODY_FEE = Column(DECIMAL(19,4)) #  托管费
    IS_QUANT = Column(TINYINT(3)) # 是否量化
    IS_HEDGING = Column(TINYINT(3)) #  是否对冲
    HERMESFIRSTTIME = Column(TIMESTAMP) # 记录数据首次写入时间
    HERMESUPTIME = Column(TIMESTAMP) # 记录数据更新时间
    

class PfundInstInfo(Base):
    '''投顾机构信息'''
        
    __tablename__ = 'pfund_inst_info'                        

    ID = Column(BIGINT, primary_key=True) # 信息编码

    PARTY_ID = Column(BIGINT(20)) # 机构内部编码
    PROFILE = Column(TEXT) # 公司简介
    IDEA_STRATEGY= Column(TEXT) # 投资理念
    KEY_PERSON= Column(TEXT) # 核心人物
    UPDATE_TIME= Column(DATETIME) # 更新时间
    TMSTAMP= Column(BIGINT(20)) # 后台系统修改数据时间戳
    RECORD_DATE= Column(DATE) # 登记日期
    REG_CD= Column(CHAR(20)) # 登记编码
    LEGAL_REP= Column(CHAR(100)) # 法定代表人/执行事务合伙人(委派代表)
    IS_QUALIFIED= Column(TINYINT(3)) # 法人是否有从业资格
    QUALIFY_WAY= Column(CHAR(20)) # 法人从业资格认定方式
    EMP_NUM= Column(Integer) # 员工数量
    MAIN_FUND_TYPE= Column(CHAR(100)) # 管理基金主要类别
    PRIVATE_IND_SCALE= Column(CHAR(20)) # 私募证券基金自主发行规模
    PRIVATE_CON_SCALE= Column(CHAR(20)) # 私募证券基金顾问管理规模
    PE_SCALE= Column(CHAR(20)) # 私募股权基金规模
    VC_SCALE= Column(CHAR(20)) # 创业投资基金规模
    OTHER_SCALE= Column(CHAR(20)) # 其他私募基金规模
    IS_FUND_MANA= Column(BIGINT) # 是否公募基金管理人
    PAA_SCALE= Column(CHAR(20)) # 私募资产配置基金规模
    RECORD_STATUS= Column(CHAR(20)) # 备案状态
    MANAGEMENT_SCALE= Column(CHAR(100)) # 管理规模区间
    HERMESFIRSTTIME= Column(TIMESTAMP) # 记录数据首次写入时间
    HERMESUPTIME= Column(TIMESTAMP) # 记录数据更新时间




from typing import Union
import datetime

from ..api.raw import RawDataApi
from ..api.basic import BasicDataApi
from ..api.derived import DerivedDataApi
from ..api.view import ViewDataApi


class DataManager:

    def __init__(self, start_time, end_time):
        # start_time: None, datetime or '20100101'
        # end_time:   None, datetime or '20100101'
        self.start_time = start_time if isinstance(start_time, datetime.datetime) else datetime.datetime.strptime(start_time,'%Y%m%d')
        self.end_time = end_time if isinstance(end_time, datetime.datetime) else datetime.datetime.strptime(end_time,'%Y%m%d')
        if self.end_time.date() == datetime.datetime.now().date():
            self.end_time = self.end_time - datetime.timedelta(days=1)

    def init(self):
        pass

    def terminate(self):
        pass

    @property
    def start_date(self):
        return self.start_time.date()

    @property
    def end_date(self):
        return self.end_time.date()

    @staticmethod
    def _db_api_helper(api_object: Union[RawDataApi, BasicDataApi, DerivedDataApi, ViewDataApi], func_name: str, *args, **kwargs):
        if func_name.startswith("__"):
            print(f'should not call func that starts with "__" (name){func_name}')
            return

        try:
            func = getattr(api_object, func_name)
        except AttributeError as e:
            print(e)
            return

        if not callable(func):
            print(f'{func_name} is not callable')
            return
        return func(*args, **kwargs)

    # 读取 raw db 中的数据
    @staticmethod
    def raw_data(func_name: str, *args, **kwargs):
        return DataManager._db_api_helper(RawDataApi(), func_name, *args, **kwargs)

    # 读取 basic db 中的数据
    @staticmethod
    def basic_data(func_name: str, *args, **kwargs):
        return DataManager._db_api_helper(BasicDataApi(), func_name, *args, **kwargs)

    # 读取 derived db 中的数据
    @staticmethod
    def derived_data(func_name: str, *args, **kwargs):
        return DataManager._db_api_helper(DerivedDataApi(), func_name, *args, **kwargs)

    # 读取 view db 中的数据
    @staticmethod
    def view_data(func_name: str, *args, **kwargs):
        return DataManager._db_api_helper(ViewDataApi(), func_name, *args, **kwargs)



from .cas_connection import CassandraConnector
from ...view.cas.fund_estimated_nav import FundEstimatedNav
from ...view.cas.realtime_index_price import RealtimeIndexPrice
from ...view.cas.realtime_index_price_snapshot import RealtimeIndexPriceSnapshot
from ...view.cas.tag_method_index_level import TagMethodIndexLevel
from ...view.cas.third_party_portfolio_nav import ThirdPartyPortfolioNav
from cassandra.cqlengine.management import sync_table, drop_table

CassandraConnector().get_conn()


def sync_cas_table():
    sync_table(FundEstimatedNav)
    sync_table(RealtimeIndexPrice)
    sync_table(RealtimeIndexPriceSnapshot)
    sync_table(TagMethodIndexLevel)
    sync_table(ThirdPartyPortfolioNav)


def drop_cas_table():
    drop_table(FundEstimatedNav)
    drop_table(RealtimeIndexPrice)
    drop_table(RealtimeIndexPriceSnapshot)
    drop_table(ThirdPartyPortfolioNav)


if __name__ == '__main__':
    sync_cas_table()
    # drop_cas_table()



import boto3
import sys
import pandas
sys.path.append('/Users/cjiang/Documents/GitHub/surfing/python/')
from surfing.util.singleton import Singleton
from surfing.util.config import SurfingConfigurator


class S3Connector(metaclass=Singleton):

    def __init__(self, config_path=None) -> None:
        if config_path:
            self.settings = SurfingConfigurator(config_path).get_aws_settings()
        else:
            self.settings = SurfingConfigurator().get_aws_settings()

    def get_conn(self):
        return boto3.resource(
            's3',
            region_name=self.settings.region_name,
            aws_access_key_id=self.settings.aws_access_key_id,
            aws_secret_access_key=self.settings.aws_secret_access_key,
        )

    def get_client(self):
        return boto3.client(
            's3',
            region_name=self.settings.region_name,
            aws_access_key_id=self.settings.aws_access_key_id,
            aws_secret_access_key=self.settings.aws_secret_access_key,
        )


    def put_obj(self, obj, bucket, key):
        client = self.get_client()
        client.put_object(Body=obj, Bucket=bucket, Key=key)

    def get_obj(self, bucket, key):
        client = self.get_client()
        obj = client.get_object(Bucket=bucket, Key=key)
        return obj['Body'].read()

    def put_df(self, df, bucket, key):
        path = f's3://{bucket}/{key}'
        df.to_feather(path)

    def get_df(self, bucket, key):
        path = f's3://{bucket}/{key}'
        return pandas.read_feather(path)

    def put_private_obj(self, obj, key):
        self.put_obj(obj, self.settings.research_private_bucket, key)

    def put_public_obj(self, obj, key):
        self.put_obj(obj, self.settings.research_public_bucket, key)

    def get_private_obj(self, key):
        return self.get_obj(self.settings.research_private_bucket, key)

    def get_public_obj(self, key):
        return self.get_obj(self.settings.research_public_bucket, key)

    def put_public_df(self, df, key):
        self.put_df(df, self.settings.research_public_bucket, key)

    def put_private_df(self, df, key):
        self.put_df(df, self.settings.research_private_bucket, key)

    def get_public_df(self, key):
        return self.get_df(self.settings.research_public_bucket, key)

    def get_private_df(self, key):
        return self.get_df(self.settings.research_private_bucket, key)

    # ----------
    def upload_public_file(self, file_path, file_key):
        s3 = self.get_conn()
        s3.meta.client.upload_file(file_path, self.settings.research_public_bucket, file_key, ExtraArgs={'ContentType': 'text/plain'})

'''
buckets = s3_client.list_buckets()
for bucket in buckets['Buckets']:
    print(bucket['CreationDate'].ctime(), bucket['Name'])
'''

if __name__ == '__main__':
    print(SurfingConfigurator('../../../etc/config.json').get_aws_settings())
    s3 = S3Connector().get_conn()
    s3_client = S3Connector().get_client()
    to_insert = b'this is a test'
    S3Connector().put_private_obj(to_insert, 'test1.txt')
    a = S3Connector().get_private_obj('test1.txt')
    print(type(a), a)
